/**********************************************************
 * 
 * GenericSingleLinkedList:
 *
 *		Collection class that implements
 *		a single linked list data structure
 *		using generics.
 * 
 *      
 *
 *********************************************************/
using System;
using System.Collections;
using System.Collections.Generic; // for Generics

namespace Chapter_17
{
    // linked list items are
    // managed via nodes
	public class Node<T>
	{
		public T data;
		public Node<T> next;
	}

    public struct NodeStruct<T>
    {
        public T data;
        public Node<T> next;
    }

	/// <summary>
	/// Uses System.Collection.Generic Interfaces to Create
	/// a Single Linked List Data Structure.
	/// </summary>
	public class GenericSingleLinkedList<T>: IList<T>, IEnumerable
	{
		// refers to first Node<T> in list
		Node<T> head = null;

		// used to sync this list
		private static object syncObj;

		/// <summary>
		/// Instantiates static fields
		/// </summary>
		static GenericSingleLinkedList()
		{
			// init object to sync this list with
			syncObj = new object();
		}

		/// <summary>
		/// Not used
		/// </summary>
		public GenericSingleLinkedList()
		{
		}

        #region IList<T> Members

 		/// <summary>
		/// Returns position in list of specified object
		/// </summary>
		/// <param name="value">object to look for</param>
		/// <returns>
		/// zero based position when found
		/// -1 when not found
		/// </returns>
        public int IndexOf(T item)
        {
            // keep track of location
            Node<T> curr = head;
            int count = 0;

            // look for object until found
            //	or reached end-of-list
            while (curr != null && !(curr.data.Equals(item)))
            {
                curr = curr.next;
                count++;
            }

            // end of list?
            if (curr != null)
            {
                // return position
                return count;
            }

            // indicate object not found
            return -1;
        }

        /// <summary>
        /// Add a Node<T> at the specified postion
        /// </summary>
        /// <param name="index">position to add Node<T></param>
        /// <param name="item">data Node<T> will reference</param>
        public void Insert(int index, T item)
        {
            // inserting to an empty list
            if (head == null)
            {
                // can only add to the
                //	first element of an
                //	empty list
                if (index > 0)
                {
                    throw new ArgumentOutOfRangeException("index", index, "Selected index greater than list size.");
                }

                // create the Node<T> and
                //	init it's data
                head = new Node<T>();
                head.data = item;
                head.next = null;
            }
            else
            {
                // keep track of location
                int count = 0;
                Node<T> prev = head;
                Node<T> curr = head;

                // keep going until position
                //	found or end of list
                while (count < index && curr.next != null)
                {
                    prev = curr;
                    curr = curr.next;
                    count++;
                }

                count++;

                // trying to insert beyond
                //	end of list
                if (index > count)
                {
                    throw new ArgumentOutOfRangeException("index", index, "Selected index greater than list size.");
                }

                // insert at end of list
                if (curr.next == null)
                {
                    curr.next = new Node<T>();
                    curr.next.data = item;
                    curr.next.next = null;
                }
                else	// insert in middle of list
                {
                    prev.next = new Node<T>();
                    prev.next.data = item;
                    prev.next.next = curr;
                }
            }
        }

		/// <summary>
		/// Deletes a Node<T> at a specified position
		/// </summary>
		/// <param name="index">Position to delete Node<T></param>
        public void RemoveAt(int index)
        {
            // it doesn't make sense to try to 
            //	remove from an empty list
            if (head == null)
            {
                throw new ArgumentOutOfRangeException("index", index, "Single Linked List is empty.");
            }

            // keep track of position
            int count = 0;

            // deleting first Node<T>?
            if (count == index)
            {
                head = head.next;
            }
            else	// look for Node<T> at index
            {
                // keep track of current Node<T>
                //	and the one before it
                Node<T> prev = head;
                Node<T> curr = head;

                // keep looking until we get to
                //	the Node<T> at index
                while (count < index)
                {
                    // if where in here and get to the
                    //	end of list, there is a problem
                    if (curr.next == null)
                    {
                        throw new ArgumentOutOfRangeException("index", index, "Selected index greater than list size.");
                    }

                    // adjust previous and
                    //	current Node<T> refs
                    prev = curr;
                    curr = curr.next;

                    // keep track of location
                    count++;
                }

                // Node<T> found - delete by removing
                //	the Node<T> from the list
                prev.next = curr.next;
            }
        }

        /// <summary>
        /// Retrieves and sets info by index
        /// </summary>
        public T this[int index]
        {
            get
            {
                // it doesn't make sense to try to 
                //	retrieve from an empty list
                if (head == null)
                {
                    throw new ArgumentOutOfRangeException("index", index, "Single Linked List is empty.");
                }

                // keep track of position
                int count = 0;

                // retrieving first Node<T>?
                if (count == index)
                {
                    return head.data;
                }
                else	// look for Node<T> at index
                {
                    // keep track of current Node<T>
                    Node<T> curr = head;

                    // keep looking until we get to
                    //	the Node<T> at index
                    while (count < index)
                    {
                        // if where in here and get to the
                        //	end of list, there is a problem
                        if (curr.next == null)
                        {
                            throw new ArgumentOutOfRangeException("index", index, "Selected index greater than list size.");
                        }

                        // adjust Node<T> ref
                        curr = curr.next;

                        // keep track of location
                        count++;
                    }

                    // return object
                    return curr.data;
                }
            }
            set
            {
                // it doesn't make sense to try to 
                //	modify an empty list
                if (head == null)
                {
                    throw new ArgumentOutOfRangeException("index", index, "Single Linked List is empty.");
                }

                // keep track of position
                int count = 0;

                // retrieving first Node<T>?
                if (count == index)
                {
                    head.data = value;
                }
                else	// look for Node<T> at index
                {
                    // keep track of current Node<T>
                    Node<T> curr = head;

                    // keep looking until we get to
                    //	the Node<T> at index
                    while (count < index)
                    {
                        // if where in here and get to the
                        //	end of list, there is a problem
                        if (curr.next == null)
                        {
                            throw new ArgumentOutOfRangeException("index", index, "Selected index greater than list size.");
                        }

                        // adjust Node<T> ref
                        curr = curr.next;

                        // keep track of location
                        count++;
                    }

                    // return object
                    curr.data = value;
                }
            }
        }

        #endregion

        #region ICollection<T> Members

        /// <summary>
        /// Adds object to end of list
        /// </summary>
        /// <param name="item">object to add</param>
        public void Add(T item)
        {
            // inserting to an empty list
            if (head == null)
            {
                // create the Node<T> and
                //	init it's data
                head = new Node<T>();
                head.data = item;
                head.next = null;
            }
            else
            {
                // keep track of location
                Node<T> curr = head;

                // keep going until position
                //	found or end of list
                while (curr.next != null)
                {
                    curr = curr.next;
                }

                // insert at end of list
                curr.next = new Node<T>();
                curr.next.data = item;
                curr.next.next = null;
            }
        }

        /// <summary>
        /// Removes all Node<T>s from list
        /// </summary>
        public void Clear()
        {
            // make head point to nothing
            head = null;
        }

        /// <summary>
        /// Determine if specified object is in list
        /// </summary>
        /// <param name="item">object to look for</param>
        /// <returns>
        /// true  - object is in list
        /// false - object is not in list
        /// </returns>
        public bool Contains(T item)
        {
            // keep track of location
            Node<T> curr = head;

            // look for object until found
            //	or reached end-of-list
            while (curr != null && !(curr.data.Equals(item)))
            {
                curr = curr.next;
            }

            // end of list?
            if (curr != null)
            {
                // object is in list
                return true;
            }

            // object is not in list
            return false;
        }

        /// <summary>
        /// Copies list to array at specified position
        /// </summary>
        /// <param name="array">array to copy to</param>
        /// <param name="arrayIndex">index in array to begin populating</param>
        public void CopyTo(T[] array, int arrayIndex)
        {
            // start at the beginning
            Node<T> curr = head;

            // go until either the array is full
            //	or we've run out of items to copy
            for (int i = arrayIndex; i < array.Length && curr != null; i++, curr = curr.next)
            {
                array.SetValue(curr.data, i);
            }

            // array isn't large enough for list
            if (curr != null)
            {
                throw new ArgumentException("Array size too small.", "array");
            }
        }

        /// <summary>
        /// Returns the number of items in list
        /// </summary>
        public int Count
        {
            get
            {
                // keep track of position
                Node<T> curr = head;
                int count;

                // iterate to end
                for (count = 0; curr != null; count++, curr = curr.next) ;

                // return number of items in list
                return count;
            }
        }

        /// <summary>
        /// Tells if this is a read-only list
        /// </summary>
        public bool IsReadOnly
        {
            get
            {
                // this list is read/write
                return false;
            }
        }


        /// <summary>
        /// Deletes specified object from list
        /// </summary>
        /// <param name="item">object to remove</param>
        /// <returns>true if item is member of list, otherwise false</returns>
        public bool Remove(T item)
        {
            bool itemExists = true;

            // keep track of location
            Node<T> prev = head;
            Node<T> curr = head;

            // look for object until found
            //	or reached end-of-list
            while (curr != null && !(curr.data.Equals(item)))
            {
                prev = curr;
                curr = curr.next;
            }

            // check for end of list
            if (curr != null)
            {
                // drop Node<T> from list
                prev.next = curr.next;
            }
            else
            {
                // value not found
                itemExists = false;
            }

            return itemExists;
        }

        #endregion

        #region IEnumerable<T> Members

        /// <summary>
        /// Returns enumerator.
        /// </summary>
        /// <returns>GenericSingleLinkedListEnumerator</returns>
        public IEnumerator<T> GetEnumerator()
        {
            return (IEnumerator<T>)new GenericSingleLinkedListEnumerator(head);
        }

        #endregion

        /// <summary>
        /// Enumerator for SingleLinkedList collection
        /// </summary>
        public class GenericSingleLinkedListEnumerator : IEnumerator<T>, IEnumerator
        {
            Node<T> curr;	// keeps track of current postion
            Node<T> head;	// points to first Node<T>

            // init enumerator with head of
            //	SingleLinkedList collection
            public GenericSingleLinkedListEnumerator(Node<T> head)
            {
                // set head
                this.head = head;

                // rewind position
                Reset();
            }

            public void Dispose()
            {
                // implement dispose if necessary
            }

            #region Implementation of IEnumerator

            // rewinds postion to beginning of list
            public void Reset()
            {
                // set curr to just before head
                curr = new Node<T>();
                curr.next = head;
            }

            // move to next position in list
            public bool MoveNext()
            {
                // make sure we aren't at the end of the list
                if (curr.next != null)
                {
                    // move to next postion
                    curr = curr.next;
                    return true;
                }

                // can't move, we were at the end
                return false;
            }

            // return data at the current position
            public T Current
            {
                get
                {
                    // return the data
                    return curr.data;
                }
            }
            #endregion

			#region IEnumerator Members

			object IEnumerator.Current
			{
				get { throw new NotImplementedException(); }
			}

			#endregion
		}

		#region IEnumerable Members

		IEnumerator IEnumerable.GetEnumerator()
		{
			throw new NotImplementedException();
		}

		#endregion
	}
}
